package ch.ipi.hello.web;

import static javax.ws.rs.core.Response.Status.NOT_ACCEPTABLE;

import java.util.List;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import javax.enterprise.inject.Any;
import javax.enterprise.inject.Instance;
import javax.inject.Inject;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Request;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.core.UriInfo;
import javax.ws.rs.core.Variant;
import javax.ws.rs.core.Variant.VariantListBuilder;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;

public class AbstractFormApplicationResource {

  private static final Logger LOG = LoggerFactory.getLogger(AbstractFormApplicationResource.class);
  private static final boolean VERBOSE = true;

  private static final String[] REPRESENTATIONS = { MediaType.TEXT_PLAIN + "; charset=UTF-8",
      "application/transit+json", "application/transit+json;verbose=true", "application/transit+msgpack",
      MediaType.APPLICATION_JSON };

  private static final List<Variant> REPR_VARIANTS = variants(REPRESENTATIONS);

  private static final MediaType[] REPR_TYPES;

  private static final String[] FORM_REPRESENTATIONS = { MediaType.APPLICATION_XML, "application/transit+json",
      "application/transit+json;verbose=true", "application/transit+msgpack", };

  private static final List<Variant> FORM_REPR_VARIANTS = variants(FORM_REPRESENTATIONS);

  static {
    final int n = REPR_VARIANTS.size();
    REPR_TYPES = new MediaType[n];
    for (int i = 0; i < n; i++) {
      REPR_TYPES[i] = REPR_VARIANTS.get(i).getMediaType();
    }
  }

  @Any
  @Inject
  private Instance<Object> cdiFactory;

  @CheckForNull
  private HttpServletRequest requestForSecurity;

  @PostConstruct
  private void postConstruct() {
    if (VERBOSE) {
      LOG.debug("initialize {}@{}", this.getClass().getSimpleName(),
          Integer.toHexString(System.identityHashCode(this)));
    }
  }

  /**
   * Gets a CDI factory.
   * @param <T> the requested type
   * @param type the requested type
   * @return the factory
   */
  @Nonnull
  protected <T> Instance<T> factory(final Class<T> type) {
    return this.cdiFactory.select(type);
  }

  /**
   * Sets the request. (Injected by jax-rs).
   * @param req the request
   */
  @Context
  public void setHttpServletRequest(final HttpServletRequest req) {
    this.requestForSecurity = req;
  }

  /**
   * Builds a variant list.
   * @param types the media types
   * @return the variants
   */
  @Nonnull
  protected static List<Variant> variants(final String... types) {
    final int n = types.length;
    final MediaType[] mt = new MediaType[n];
    for (int i = 0; i < n; i++) {
      mt[i] = MediaType.valueOf(types[i]);
    }
    return VariantListBuilder.newInstance().mediaTypes(mt).build();
  }

  protected WebApplicationException throwMessage(final Status status, final String msg, final Object... args) {
    throw new WebApplicationException(String.format(msg, args), status);
  }

  /**
   * The given value as response.
   * @param value the value
   * @param extraVariants extra variants to support
   * @return the response
   */
  @Nonnull
  protected ResponseBuilder representations(@Nullable final Object value, final String... extraVariants) {
    final List<Variant> vars = extraVariants.length == 0 ? REPR_VARIANTS
        : Lists.newArrayList(Iterables.concat(REPR_VARIANTS, variants(extraVariants)));
    return Response.ok(value).variants(vars);
  }

  /**
   * The given form as response.
   * @param form the form
   * @param extraVariants extra variants to support
   * @return the response
   */
  @Nonnull
  protected ResponseBuilder formRepresentations(@Nullable final Object form, final String... extraVariants) {
    final List<Variant> vars = extraVariants.length == 0 ? FORM_REPR_VARIANTS
        : Lists.newArrayList(Iterables.concat(FORM_REPR_VARIANTS, variants(extraVariants)));
    return Response.ok(form).variants(vars);
  }

  /**
   * Emits a stacktrace for debugging.
   * @param info the uri info
   */
  @GET
  @POST
  @Path("invocation-trace")
  public void showInvocationStackTrace(@Context final UriInfo info) {
    throw new RuntimeException(Lists.newArrayList(info.getMatchedResources(), info.getMatchedURIs()).toString());
  }

  /**
   * The given value as resource.
   * @param <T> the value type
   * @param value the value
   * @return the resource
   */
  @Nonnull
  protected <T> ValueAsResource<T> resource(@Nullable final T value) {
    return asResource(value, REPR_TYPES);
  }

  /**
   * The given value as resource.
   * @param <T> the value type
   * @param value the value
   * @param types the supported types
   * @return the resource
   */
  @Nonnull
  protected <T> ValueAsResource<T> asResource(@Nullable final T value, final MediaType... types) {
    return new ValueAsResource<>(value, types);
  }

  /**
   * Value as resource wrapper.
   * @param <T> the value type
   */
  protected static class ValueAsResource<T> {

    private final T value;
    private final MediaType[] types;

    /**
     * Creates this.
     * @param value the value
     * @param types the supported types
     */
    public ValueAsResource(@Nullable final T value, final MediaType... types) {
      super();
      this.value = value;
      this.types = types;
    }

    /**
     * Determines the available variants.
     * @return the variants
     */
    @Nonnull
    protected List<Variant> getVariants() {
      if (this.types.length == 0) {
        return ImmutableList.of();
      }
      return VariantListBuilder.newInstance().mediaTypes(this.types).build();
    }

    /**
     * Determines the response variant.
     * @param request the request
     * @param response the response
     * @return the response again
     */
    @Nonnull
    protected ResponseBuilder determineVariant(final Request request, final ResponseBuilder response) {
      final List<Variant> variants = getVariants();
      if (!variants.isEmpty()) {
        final Variant variant = request.selectVariant(variants);
        if (variant == null) {
          response.status(NOT_ACCEPTABLE);
        }
        response.variant(variant == null ? variants.get(0) : variant);
      }
      return response;
    }

    /**
     * Hook method for sub classes.
     * @param response the response builder
     * @return the response
     */
    @Nonnull
    protected Response buildResponse(final ResponseBuilder response) {
      return response.build();
    }

    /**
     * Gets the value.
     * @param request the request
     * @return the response
     */
    @GET
    public Response getValue(@Context final Request request) {
      return buildResponse(determineVariant(request, Response.ok(this.value)));
    }

  }

}
