package ch.ipi.hello.web;

import static javax.ws.rs.core.Response.Status.NOT_FOUND;

import java.time.LocalDate;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.UUID;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.ws.rs.GET;
import javax.ws.rs.MatrixParam;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.core.UriInfo;

@Path("problem")
public class ProblemResource {

  //  req "problem/items?msg=foo" -X POST
  //  req "problem/items;id=414ea4a5-d75f-413e-9c77-fc596022cdc5?msg=bar" -X POST
  //
  //  req "problem/items/414ea4a5-d75f-413e-9c77-fc596022cdc5"
  //
  //  ACCEPT="application/json" req "problem/items"
  //  ACCEPT="application/json" req "problem/items;pattern=foo"
  //  req "problem/items/0"
  //  req "problem/items;pattern=foo/0"
  //
  //  ACCEPT="application/json" req "problem/locateItemsWORKAROUND"
  //  ACCEPT="application/json" req "problem/locateItemsWORKAROUND;pattern=foo"
  //  req "problem/locateItemsWORKAROUND/0"
  //  req "problem/locateItemsWORKAROUND;pattern=foo/0"

  @Context
  private UriInfo uriInfo;

  @POST
  @Path("items")
  public Response create(@MatrixParam("id") final UUID id, @QueryParam("msg") final String msg) {
    final UUID regId = Repo.register(id == null ? UUID.randomUUID() : id, msgOrNow(msg));
    return Response.created(this.uriInfo.getAbsolutePathBuilder().segment(regId.toString()).build()).build();
  }

  private static String msgOrNow(final String msg) {
    return msg == null ? LocalDate.now().toString() : msg;
  }

  @Path("items")
  public ItemCollectionResource locateItems(@MatrixParam("pattern") final String pattern) {
    final ArrayList<UUID> ids = new ArrayList<>(
        pattern == null ? Repo.findIds(Collections.emptyList()) : Repo.findIds(Pattern.compile(pattern)));
    return new ItemCollectionResource(ids);
  }

  @Path("locateItemsWORKAROUND")
  public ItemCollectionResource locateItemsWORKAROUND(@MatrixParam("pattern") final String pattern) {
    final ArrayList<UUID> ids = new ArrayList<>(
        pattern == null ? Repo.findIds(Collections.emptyList()) : Repo.findIds(Pattern.compile(pattern)));
    return new ItemCollectionResource(ids);
  }

  @Path("items/{id:[A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9]-[A-Za-z0-9\\-]+}")
  public SingleItemResource locateItem(@PathParam("id") final UUID id) {

    final Collection<UUID> matching = Repo.findIds(Collections.singletonList(id));
    if (matching.size() != 1) {
      throw new WebApplicationException(Status.NOT_FOUND);
    }

    return new SingleItemResource(matching.iterator().next());
  }

  public static class ItemCollectionResource {

    private final List<UUID> ids;

    public ItemCollectionResource(final List<UUID> ids) {
      this.ids = ids;
    }

    @GET
    public List<UUID> show() {
      return this.ids;
    }

    @Path("{idx:-?[0-9]+}")
    public SingleItemResource locateInstanceAtIndex(@PathParam("idx") final int idx) {
      final int n = this.ids.size();
      if (n == 0) {
        throw new WebApplicationException("empty list, no indexing possible", NOT_FOUND);
      }
      try {
        return new SingleItemResource(this.ids.get(idx < 0 ? n + idx : idx));

      } catch (final IndexOutOfBoundsException ignored) {
        throw new WebApplicationException(String.format("valid indicies from %s to %s", -n, n - 1), NOT_FOUND);

      }
    }

  }

  public static class SingleItemResource {

    private final UUID id;

    public SingleItemResource(final UUID id) {
      this.id = id;
    }

    @GET
    public Object show() {
      return Repo.readOne(this.id);
    }

  }

  //-------------------------------------------------------

  public static class Repo {

    private static final Map<UUID, String> REPO = new LinkedHashMap<>();

    static {
      final List<UUID> rnd = Stream.generate(UUID::randomUUID).limit(10L).collect(Collectors.toList());
      rnd.stream().forEachOrdered(id-> register(id, id.toString()));
    }

    public static Collection<UUID> findIds(final List<UUID> requiredIds) {
      return requiredIds.isEmpty() ? REPO.keySet()
          : requiredIds.stream().filter(REPO::containsKey).collect(Collectors.toList());
    }

    public static List<UUID> findIds(final Pattern msgPattern) {
      return REPO.entrySet().stream().filter(e-> msgPattern.matcher(e.getValue()).matches()).map(Entry::getKey)
          .collect(Collectors.toList());
    }

    public static String readOne(final UUID uuid) {
      return REPO.get(uuid);
    }

    public static UUID register(final UUID id, final String data) {
      REPO.put(id, data);
      return id;
    }

  }

}
