package ch.ipi.hello.web.form;

import static javax.ws.rs.core.Response.Status.BAD_REQUEST;
import static javax.ws.rs.core.Response.Status.NOT_FOUND;

import java.net.URI;
import java.time.Instant;
import java.util.List;
import java.util.UUID;
import java.util.stream.Collectors;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.MatrixParam;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Request;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;
import javax.ws.rs.core.UriInfo;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.ipi.hello.web.AbstractFormApplicationResource;
import ch.ipi.hello.web.form.instance.FormInstanceResource;
import ch.ipi.hello.web.form.instance.GenericFormInstanceResource;
import ch.ipi.hello.web.form.instance.collection.FormInstanceCollectionResource;
import ch.ipi.hello.web.form.instance.collection.GenericFormInstanceCollectionResource;

import com.google.common.base.Optional;

@Path("/form")
public class GenericFormResource extends AbstractFormApplicationResource {

  private static final Logger LOG = LoggerFactory.getLogger(GenericFormResource.class);

  /**
   * Gets an appropriate {@link FormInstanceResource} for this form type.
   * @param instance the form
   * @return the resource dispatcher
   */
  @Nonnull
  public FormInstanceResource single(final FormularData instance) {
    final FormInstanceResource single = createSub(FormInstanceResource.class);
    single.setInstance(instance);
    return single;
  }

  /**
   * Gets an appropriate {@link FormInstanceResource} for this form type.
   * @param instance the form
   * @return the resource dispatcher
   */
  @Nonnull
  public FormInstanceResource single(final UUID instance) {
    final FormInstanceResource single = createSub(FormInstanceResource.class);
    single.setID(instance);
    return single;
  }

  /**
   * Gets an appropriate {@link FormInstanceCollectionResource} for this form type.
   * @param instances the forms
   * @return the resource dispatcher
   */
  @Nonnull
  public FormInstanceCollectionResource collection(final Iterable<UUID> instances) {
    final FormInstanceCollectionResource coll = createSub(FormInstanceCollectionResource.class);
    coll.setIDs(instances);
    return coll;
  }

  @Nonnull
  private <S> S createSub(final Class<S> type) {
    return factory(type).get();
  }

  /**
   * Gets the master form.
   * @return the master form
   */
  @Nonnull
  public Optional<FormularData> getMaster() {
    return Optional.absent();
  }

  /**
   * Ping-Pong.
   * @param pong the pong or {@code null}
   * @return the pong
   */
  @GET
  @Path("ping")
  public String ping(@QueryParam("pong") @Nullable final String pong) {
    return pong == null ? Instant.now().toString() : pong;
  }

  /**
   * Deliver the form home.
   * @param request the request
   * @return the response
   */
  @GET
  public Response deliverFormHome(@Context final Request request) {
    return Response.ok("form home").build();
  }

  /**
   * Locates the master instance.
   * @return the master
   */
  @Path("master")
  public GenericFormInstanceResource locateMaster() {
    throw new WebApplicationException("not implemented");
  }

  /**
   * Redirect to the master instance.
   * @param uriInfo the uri info
   * @return the redirect
   */
  @Nonnull
  public Response redirectToMaster(@Context final UriInfo uriInfo) {
    throw new WebApplicationException("not implemented");
  }

  /**
   * Locates a form instance.
   * @param id the id
   * @param state the required state or {@code null} if any
   * @return the instance
   */
  @Path("instances/{id:[A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9]-[A-Za-z0-9\\-]+}")
  public GenericFormInstanceResource locateInstance(@PathParam("id") final String id,
      @MatrixParam("state") @Nullable final String state) {

    LOG.info("locateInstance");

    final Optional<UUID> parsed = FormularIdentifiers.parseId(id);
    if (!parsed.isPresent()) {
      throwMessage(BAD_REQUEST, "invalid id: %s", id);
    }
    final UUID uuid = parsed.get();
    final FormularData instance = Service.readOne(uuid);
    if (instance == null) {
      throw throwMessage(NOT_FOUND, "%s not found", uuid);
    }
    return single(instance).getResource();
  }

  /**
   * Redirect to the form instance.
   * @param uriInfo the uri info
   * @param id the id
   * @return the redirect
   */
  @Nonnull
  public Response redirectToInstance(final UriInfo uriInfo, final UUID id) {
    throw new WebApplicationException("not implemented");
  }

  //*********************************
  // CLIENT API BREAKING WORKAROUND

  // @Path("instances/{id:[A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9]-[A-Za-z0-9\\-]+}") -->
  @Path("locateInstanceWORKAROUND/{id:[A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9][A-Za-z0-9]-[A-Za-z0-9\\-]+}")
  public GenericFormInstanceResource locateInstanceWORKAROUND(@PathParam("id") final String id,
      @MatrixParam("state") @Nullable final String state) {
    return locateInstance(id, state);
  }

  //@Path("instances") ->
  @Path("locateInstancesWORKAROUND")
  public GenericFormInstanceCollectionResource locateInstancesWORKAROUND(
      @MatrixParam("state") @Nullable final String state, @MatrixParam("id") final List<String> ids,
      @MatrixParam("filterValue") final List<String> valueFilterExpressions) {
    return locateInstances(state, ids, valueFilterExpressions);
  }

  //@Path("instances") -->
  @Path("createFormInstanceWORKAROUND")
  @POST
  public Response createFormInstanceWORKAROUND(@MatrixParam("id") final List<String> id,
      @MatrixParam("state") @Nullable final String state, @QueryParam("state") @Nullable final String stateQueryParam,
      @QueryParam("master") @DefaultValue("false") final boolean createMaster, @Context final UriInfo uriInfo,
      @QueryParam("redirect") @DefaultValue("false") final boolean redirect) {
    return createFormInstance(id, state, stateQueryParam, createMaster, uriInfo, redirect);
  }

  //*********************************

  /**
   * Locates form instances.
   * @param state the required state or {@code null} if any
   * @param ids the required ids
   * @param valueFilterExpressions requirements on the values of the forms
   * @return the instances
   */
  @Path("instances")
  public GenericFormInstanceCollectionResource locateInstances(@MatrixParam("state") @Nullable final String state,
      @MatrixParam("id") final List<String> ids,
      @MatrixParam("filterValue") final List<String> valueFilterExpressions) {

    LOG.info("locateInstances");

    final List<UUID> requiredIds = ids.stream().map(FormularIdentifiers.PARSE).collect(Collectors.toList());
    return collection(Service.findFormInstanceIds(requiredIds)).getResource();
  }

  /**
   * Creates a form instance.
   * @param id the requested id or {@code null} for a generated id
   * @param state the initial state or {@code null} for the first start state of the state machine
   * @param stateQueryParam additional means to give the initial state (to be consistent with POST and PUT on the
   * instance resource)
   * @param createMaster if {@code true} no state is set and a master form is created
   * @param uriInfo the uri info
   * @param redirect if {@code true}, respond with a redirect instead a created
   * @return the HTTP 201 created response
   */
  @POST
  @Path("instances")
  public Response createFormInstance(@MatrixParam("id") final List<String> id,
      @MatrixParam("state") @Nullable final String state, @QueryParam("state") @Nullable final String stateQueryParam,
      @QueryParam("master") @DefaultValue("false") final boolean createMaster, @Context final UriInfo uriInfo,
      @QueryParam("redirect") @DefaultValue("false") final boolean redirect) {

    LOG.info("createFormInstance");

    final Optional<UUID> requestedID = requestedID(id);
    final Optional<String> requestedStateQueryParam = Optional.fromNullable(stateQueryParam);
    final Optional<String> requestedState = Optional.fromNullable(state).or(requestedStateQueryParam);
    if (requestedStateQueryParam.isPresent() && !requestedState.equals(requestedStateQueryParam)) {
      throw throwMessage(BAD_REQUEST, "if both query and matrix params for state are given they must match");
    }
    if (requestedState.isPresent() && createMaster) {
      throw throwMessage(BAD_REQUEST, "either give state or master flag but not both");
    }

    final FormularData created = Service.create(fillValues(requestedID, requestedState, createMaster));

    final UUID createdId = created.getID();
    final URI tgt = uriInfo.getAbsolutePathBuilder().segment("{id}").build(createdId);
    final ResponseBuilder response = redirect ? Response.seeOther(tgt) : Response.created(tgt);
    return response.build();
  }

  private FormularData fillValues(final Optional<UUID> requestedID, final Optional<String> requestedState,
      final boolean createMaster) {
    final FormularData f = requestedID.isPresent() ? new FormularData(requestedID.get()) : new FormularData();

    if (requestedState.isPresent()) {
      f.add(requestedState.get());
    }
    return f;
  }

  private Optional<UUID> requestedID(final List<String> ids) {
    final List<UUID> parsed = ids.stream().map(FormularIdentifiers.PARSE).collect(Collectors.toList());
    final int n = parsed.size();
    if (n > 1) {
      throw throwMessage(BAD_REQUEST, "at most one ID allowed");
    }
    if (n == 1) {
      return Optional.of(parsed.get(0));
    }
    return Optional.absent();
  }

}
