package ch.ipi.hello.web.form.instance;

import java.net.URI;
import java.util.List;
import java.util.UUID;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Request;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;
import javax.ws.rs.core.UriInfo;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.ipi.hello.web.form.FormularData;
import ch.ipi.hello.web.form.Service;

import com.google.common.collect.Lists;

public class GenericFormInstanceResource extends AbstractFormInstanceResource<UUID> implements FormInstanceResource {

  private static final UUID MASTER_UUID = UUID.randomUUID();

  private static final Logger LOG = LoggerFactory.getLogger(GenericFormInstanceResource.class);

  /**
   * {@inheritDoc}
   */
  @Override
  public GenericFormInstanceResource getResource() {
    return this;
  }

  @Override
  public List<UUID> getIDs() {
    return Lists.newArrayList(getID());
  }

  /**
   * Gets the id.
   * @return the id
   */
  @Nonnull
  public UUID getID() {
    return getScope();
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void setID(final UUID instance) {
    setScope(instance);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void setInstance(final FormularData instance) {
    setScope(instance.getID());
  }

  /**
   * Redirect to this form instance.
   * @param uriInfo the uri info
   * @return the redirect
   */
  @GET
  @POST
  @Path("resolve")
  public Response redirectToInstance(@Context final UriInfo uriInfo) {
    return getForm().redirectToInstance(uriInfo, getID());
  }

  /**
   * Redirect to this form master instance.
   * @param uriInfo the uri info
   * @return the redirect
   */
  @GET
  @POST
  @Path("master")
  public Response redirectToMaster(@Context final UriInfo uriInfo) {
    return getForm().redirectToMaster(uriInfo);
  }

  /**
   * Shows the instance.
   * @param uriInfo the URI Info
   * @param withStylesheet add a stylesheet (default: {@code true})
   * @param stylesheetURI the stylesheet URI or let {@code null} for the default
   * @return the response
   */
  @GET
  public Response showInstance(@Context final UriInfo uriInfo,
      @QueryParam("stylesheet") @DefaultValue("true") final boolean withStylesheet,
      @QueryParam("stylesheetURI") @Nullable final URI stylesheetURI) {

    final List<String> f = Service.readOne(getID());
    final ResponseBuilder response = formRepresentations(f);
    return response.build();
  }

  /**
   * Merge the instance.
   * @param request the request
   * @param formData the request body or {@code null} for an empty form
   * @param state the requested state or {@code null} to extract from body or leave it at current
   */
  @PUT
  public void replaceInstance(@Context final Request request, @Nullable final FormularData formData,
      @QueryParam("state") @Nullable final String state) {
    updateInstance(false, request, formData, state);
  }

  /**
   * Replace the instance.
   * @param request the request
   * @param formData the request body or {@code null} for an empty form
   * @param state the requested state or {@code null} to extract from body or leave it at current
   */
  @POST
  public void mergeInstance(@Context final Request request, @Nullable final FormularData formData,
      @QueryParam("state") @Nullable final String state) {
    updateInstance(true, request, formData, state);
  }

  /**
   * Locates the master instance.
   * @return the master
   */
  @Path("master")
  public GenericFormInstanceResource locateMaster() {
    final GenericFormInstanceResource r = factory(GenericFormInstanceResource.class).get();
    r.setID(MASTER_UUID);
    return r;
  }

  private void updateInstance(final boolean b, final Request request, final FormularData formData, final String state) {
    LOG.info("update {} {} {}", request, formData, state);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public Response delete() {
    LOG.info("delete");
    return Response.ok("deleted").build();
  }

}
